from typing import List, Union
from cxxheaderparser.types import Parameter, DecoratedType
from re import sub

from ... import type_checks

from ..struct_backend import StructBackend

CLASS_BLACKLIST = [
    "PntosMediator",
    "PntosKeyValueStore",
    "PntosRegistry",
    "PntosStandardInertialMechanization",
]

pntos_class_template = """
/**
 * A wrapper object that implements the C API but calls into an underlying
 * Python object for the actual implementation. Facilitates the
 * implementation of a C API by Python code.
 */
typedef struct {class_name}_PyAdapter {{
    {class_name} base;
	// The Python object that actually implements the functionality
	PyObject* python_object;
}} {class_name}_PyAdapter;

/**
 * Destructor that gets called when there are no remaining references to the
 * A {class_name}_PyAdapter struct.
 */
static void cleanup_{class_name}_PyAdapter(void* memory) {{
    py::gil_scoped_acquire gil;
    {class_name}_PyAdapter* self = ({class_name}_PyAdapter*) memory;
    {destructor_body}
    if (self->python_object && self->python_object != Py_None) {{
        py::handle(self->python_object).dec_ref();
        self->python_object = NULL;
    }}

    return;
}}
{function_impls}
{ctor}
"""

function_impl_template = """\
/**
 * A trampoline, which redirects all implementation details to the
 * `PyObject* python_object` on the adapter struct.
 */
static {return_type} {class_name}_{func_name}(
    {params}) {{

    py::gil_scoped_acquire gil;
    {class_name}_PyAdapter* adapter = ({class_name}_PyAdapter*)self;
    {func_body}
    {call_prefix}py::handle(adapter->python_object).attr("{func_name}")({python_params});
    {return_statement}
}}
"""

ctor_impl_template = """\
/**
 * Constructs a new implementation of a pntOS C struct, delegating all work to
 * the passed in python object.
 *
 * Assumes GIL is acquired, only call from Python (or acquire GIL first).
 */
shared_ptr<{type_name}> create_{type_name} (py::object python_object) {{

    {type_name}_PyAdapter* out =
        pntos_new_full({type_name}_PyAdapter, cleanup_{type_name}_PyAdapter);

    {field_assignments}
    out->python_object = python_object.release().ptr();
    return shared_ptr<{type_name}>(&out->base, pntos_deleter);
}}
"""


def class_to_enum(class_name: str) -> str:
    """
    Accepts a string to a pntOS plugin struct and returns the equivalent enum
    from PntosPluginTypes.
    """
    return sub(r'(?<!^)(?=[A-Z])', '_', class_name).upper()


class PybindCToPy(StructBackend):
    def __init__(self, class_name: str):
        self.class_name = class_name
        self.ctor_buf: List[str] = []
        self.funcs_buf: List[str] = []
        self.dtor_body_buf: List[str] = []
        self.is_managed = class_name.startswith("Pntos")

    def _generate_ctor(self):
        return ctor_impl_template.format(
            type_name=self.class_name,
            field_assignments="\n    ".join(self.ctor_buf),
        )

    def _generate_dtor(self):
        return "\n    ".join(self.dtor_body_buf)

    def generate(self) -> str:
        ctor = self._generate_ctor()
        dtor = self._generate_dtor()
        if self.class_name in CLASS_BLACKLIST:
            print(f"    Ignoring blacklisted class {self.class_name}")
            return f"/* WIP: Ignoring blacklisted class {self.class_name} */"
        elif self.is_managed:
            return pntos_class_template.format(
                class_name=self.class_name,
                function_impls="\n".join(self.funcs_buf),
                ctor=ctor,
                destructor_body=dtor,
            )
        print(f"    Ignoring unmanaged class {self.class_name}")
        return f"/* WIP: Ignoring unmanaged class {self.class_name} */"

    def process_func_ptr_field_with_self(
        self,
        field_name: str,
        params: List[Parameter],
        return_t: DecoratedType,
        doc_string: str,
    ):

        return_type: str
        if type_checks.is_void(return_t):
            return_type = "void"
            call_prefix = ""
            return_statement = "return;"
        elif str_identifier := type_checks.is_char_pointer(return_t):
            return_type = "char *"
            call_prefix = "py::object out = "
            return_statement = "return g_strdup(py::cast<std::string>(out).c_str()); // WIP: leaks memory"
        elif ptr_type_identifier := type_checks.is_pointer(return_t):
            return_type = ptr_type_identifier + "*"
            call_prefix = "py::object out_py = "
            if field_name == 'clone':
                return_statement = f"""
                    if (out_py.is(py::none())) return NULL;
                    // Forward declare function.
                    shared_ptr<{ptr_type_identifier}> create_{ptr_type_identifier} (py::object python_object);
                    auto out = create_{ptr_type_identifier}(out_py);
                    pntos_ref(out.get());
                    return out.get();
                """
            elif ptr_type_identifier == 'PntosMatrix':
                return_statement = f"""
                    if (out_py.is(py::none())) return NULL;
                    auto out = numpy_to_pntos_matrix(out_py);
                    pntos_ref(out);
                    return out;
                """
            else:
                return_statement = f"""
                    if (out_py.is(py::none())) return NULL;
                    auto out = py::cast<shared_ptr<{ptr_type_identifier}>>(out_py);
                    pntos_ref(out.get());
                    return out.get();
                """

        elif value_identifier := type_checks.is_value(return_t):
            return_type = value_identifier
            call_prefix = "py::object out = "
            return_statement = f"return py::cast<{return_type}>(out);"
        else:
            return_type = "/* WIP: Unsupported return type */"
            call_prefix = ""
            return_statement = "/* return WIP unknown */;"
            print(
                f"    Generation failed for function {field_name}: unknown return type"
            )

        self.ctor_buf += [
            f"out->base.{field_name} = {self.class_name}_{field_name};"
        ]

        param_buf: List[str] = []
        body_buf: List[str] = []
        for param in params:
            param_failed_gen = False

            # Special case: char**, interpreted as an array of C
            # strings with unknown length
            if type_name := type_checks.is_string_array(param.type):
                param_buf += [f"char** {param.name}"]
                body_buf += [
                    f"""
                    std::vector<std::string> {param.name}_prep;
                    size_t ii = 0;
                    while (true) {{
                        if ({param.name}[ii] == NULL) break;
                        {param.name}_prep.push_back(std::string({param.name}[ii]));
                        ii++;
                    }}
                    """
                ]
                param_failed_gen = True

            # Special case: char*, interpreted as a C string
            elif type_name := type_checks.is_char_pointer(param.type):
                param_buf += [f"char* {param.name}"]
                body_buf += [
                    f"""
                    std::string {param.name}_prep = std::string({param.name});
                    """
                ]

            # Special case: double*, interpreted as an array of doubles
            # with unknown length
            elif type_name := type_checks.is_pointer(param.type, ["double"]):
                param_buf += [f"double * {param.name}"]
                body_buf += [
                    f"/* WIP: double array unsupported {param.name}*/ "
                ]
                param_failed_gen = True

            # A pointer to a struct starting with Pntos, that is not a special case
            # above
            elif type_name := type_checks.is_pntos_struct_pointer(param.type):
                param_buf += [f"{type_name}* {param.name}"]

                if param.name != "self":
                    if (
                        param.type.ptr_to.typename.segments[0].name
                        == 'PntosMatrix'
                    ):
                        body_buf += [
                            f"""\
                            auto {param.name}_prep = pntos_matrix_to_numpy({param.name});\
                            """
                        ]

                    elif (
                        param.type.ptr_to.typename.segments[0].name
                        == 'PntosAuxData'
                    ):
                        body_buf += [
                            f"auto {param.name}_prep = extract_aux_data({param.name});"
                        ]

                    elif type_name == 'PntosPluginList':
                        body_buf += [
                            f"""\
                            std::vector<std::shared_ptr<PntosCommonPlugin>> {param.name}_prep;
                            for (size_t ii = 0; ii < {param.name}->num_plugins; ii++) 
                                {param.name}_prep.push_back(std::shared_ptr<PntosCommonPlugin>({param.name}->plugins[ii], pntos_deleter));
                            """
                        ]

                    else:
                        body_buf += [
                            f"""\
                            shared_ptr<{type_name}> {param.name}_prep = std::shared_ptr<{type_name}>({param.name}, pntos_deleter);\
                            """
                        ]

            # A pointer to pointer to Pntos type, interpreted as an array of
            # Pntos struct pointers with unknown length
            elif type_name := type_checks.is_pntos_struct_array_pointer(
                param.type
            ):
                param_buf += [f"{type_name}** {param.name}"]
                body_buf += [
                    f"""
                    /* WIP: struct array ptr unsupported {param.name}*/
                    (void){param.name};
                    int {param.name}_prep = 0; // Generation failed
                    """
                ]
                param_failed_gen = True

            # Generic pass-by-value, which doesn't match any of the above
            # cases. Assumed this case is catching only simple primitives /
            # structs of primitives, but this assumption is probably invalid.
            # In the future we need a tighter check for values of simple types.
            elif type_name := type_checks.is_value(param.type):
                if type_name == "va_list":
                    param_buf += [
                        f"/* WIP: va_list unsupported for '{param.name}' */ {type_name} "
                    ]
                    body_buf += [
                        f"""
                        int {param.name}_prep = 0; // Generation failed
                        """
                    ]
                    param_failed_gen = True
                else:
                    param_buf += [f"{type_name} {param.name}"]
                    body_buf += [
                        f"""\
                        {type_name} &{param.name}_prep = {param.name};\
                        """
                    ]

            else:
                param_buf += [f"/* WIP: unknown '{param.name}'*/ void*"]
                param_failed_gen = True

            if param_failed_gen:
                print(
                    f"    Generation failed for function {field_name} parameter {param.name}"
                )

        python_params_buf = []
        for param in params[1:]:
            python_params_buf += [f"{param.name}_prep"]

        params_str = ",\n    ".join(param_buf)
        python_params_str = ",\n        ".join(python_params_buf)
        body_str = "\n    ".join(body_buf)

        if len(params) > 0:
            python_params_str = "\n        " + python_params_str

        self.funcs_buf += [
            function_impl_template.format(
                return_type=return_type,
                class_name=self.class_name,
                func_name=field_name,
                params=params_str,
                func_body=body_str,
                python_params=python_params_str,
                call_prefix=call_prefix,
                return_statement=return_statement,
            )
        ]

    def process_data_pointer_field(
        self,
        field_name: str,
        type_name: str,
        data_len: Union[str, int],
        doc_string: str,
        deref="",
    ):
        if isinstance(data_len, str):
            length_field = data_len.split('self->')[1]
            length_value = (
                f'python_object.attr("{length_field}").cast<size_t>()'
            )
        else:
            length_value = f'{data_len}'
        self.ctor_buf += [
            f"""
            size_t {field_name}_length = {length_value};
            auto {field_name}_py_array = python_object.attr("{field_name}").cast<py::array_t<{type_name}>>();
            auto {field_name}_pointer = {field_name}_py_array.data();
            for (size_t ii = 0; ii < {field_name}_length; ii++)
                out->base.{field_name}[ii] = {field_name}_pointer[ii];
            """
        ]

    def process_outer_managed_pointer_field(
        self, field_name: str, field_type_name: str, doc_string: str
    ):
        if field_type_name == 'PntosStringArray':
            self.ctor_buf += [
                f"""
                // Auto-convert from Python type to closest C++ type.
                auto {field_name}_str_vec = python_object.attr("{field_name}").cast<std::vector<std::string>>();

                // Convert from C++ type to C type.
                auto {field_name}_length = {field_name}_str_vec.size();
                char** {field_name}_char_array = (char**)malloc(({field_name}_length + 1) * sizeof(char*));
                for (size_t ii = 0; ii < {field_name}_length; ii++) {{
                    {field_name}_char_array[ii] = g_strdup({field_name}_str_vec[ii].c_str());
                }}
                {field_name}_char_array[{field_name}_length] = NULL;

                // Convert from C type to pntOS type.
                auto {field_name}_pntos_string_array = pntos_string_array_new({field_name}_char_array);
                free({field_name}_char_array);

                out->base.state_block_labels = {field_name}_pntos_string_array;
                """
            ]
        elif field_type_name == 'PntosMatrix':
            self.ctor_buf += [
                f"""\
                out->base.{field_name} = numpy_to_pntos_matrix(python_object.attr("{field_name}").cast<py::array_t<double>>());
                """
            ]
        else:
            self.ctor_buf += [
                f"""\
                out->base.{field_name} = python_object.attr("{field_name}").cast<{field_type_name}*>();
                """
            ]

    def process_outer_managed_pointer_array_field(
        self,
        field_name: str,
        field_type_name: str,
        data_len: Union[str, int],
        doc_string: str,
        deref="",
    ):
        pass

    def process_string_field(self, field_name: str, doc_string: str):
        self.ctor_buf += [
            f"""\
            out->base.{field_name} = g_strdup(python_object.attr("{field_name}").cast<std::string>().c_str());
            """
        ]

    def process_string_array_field(self, field_name: str, doc_string: str):
        self.ctor_buf += [
            f"""\
            auto {field_name}_prep = python_object.attr("{field_name}").cast<std::vector<std::string>>();
            auto {field_name}_length = {field_name}_prep.size();
            char** {field_name}_prep_array = (char**)malloc(({field_name}_length + 1) * sizeof(char*));
            for (size_t ii = 0; ii < {field_name}_length; ii++) {{
                {field_name}_prep_array[ii] = g_strdup({field_name}_prep[ii].c_str());
            }}
            {field_name}_prep_array[{field_name}_length] = NULL;
            out->base.{field_name} = {field_name}_prep_array;
            """
        ]
        self.dtor_body_buf += [
            f"""
            if (self->base.{field_name}) {{
                g_free(self->base.{field_name});
                self->base.{field_name} = NULL;
            }}"""
        ]

    def process_simple_field(
        self, field_name: str, field_type_name: str, doc_string: str
    ):
        self.ctor_buf += [
            f"""\
                out->base.{field_name} = python_object.attr("{field_name}").cast<{field_type_name}>();
                """
        ]

    def process_pntos_inheritance_field(
        self, field_name: str, field_type_name: str, doc_string: str
    ):
        if field_type_name == "PntosCommonPlugin":
            self.dtor_body_buf += [
                """
                if (self->base.common.identifier) {
                    g_free(self->base.common.identifier);
                    self->base.common.identifier = NULL;
                }"""
            ]
            self.ctor_buf += [
                f"""\
                std::string identifier = py::str(python_object.attr("identifier"));
                out->base.common.plugin_type = {class_to_enum(self.class_name)};
                out->base.common.identifier = g_strdup(identifier.c_str());

                out->base.common.init_plugin = {self.class_name}_init_plugin;
                out->base.common.shutdown_plugin =
                {self.class_name}_shutdown_plugin;
                """
            ]
            self.funcs_buf += [
                f"""\
                void {self.class_name}_init_plugin(struct PntosCommonPlugin* self,
                                    char* plugin_location,
                                    PntosMediator* PNTOS_NULLABLE mediator) {{
                    py::gil_scoped_acquire gil;
                    {self.class_name}_PyAdapter* adapter = ({self.class_name}_PyAdapter*)self;
                    py::handle(adapter->python_object).attr("init_plugin")(std::string(plugin_location),
                    shared_ptr<PntosMediator>(mediator, pntos_deleter));
                }}

                void {self.class_name}_shutdown_plugin(struct PntosCommonPlugin* self) {{
                    py::gil_scoped_acquire gil;
                    {self.class_name}_PyAdapter* adapter = ({self.class_name}_PyAdapter*)self;
                    py::handle(adapter->python_object)
                        .attr("shutdown_plugin")();
                }}
                """
            ]
        elif field_type_name == "PntosCommonSensorModel":
            self.ctor_buf += [
                """\
                out->base.common.engine_type = python_object.attr("engine_type").cast<PntosFusionType>();
                out->base.common.identifier = python_object.attr("identifier").cast<uint32_t>();
                """
            ]
        else:
            self.ctor_buf += [
                f"""\
                // WIP out->base.common.{field_name} = ?;
                // WIP {field_type_name} is unsupported.
                """
            ]
            print(
                f"    WIP: Ignoring unknown field common.'{field_name}' in {self.class_name}"
            )

    def process_class_docstring(self, doc_string: str):
        pass
