from typing import List, Union
from cxxheaderparser.types import Parameter, DecoratedType

from ..struct_backend import StructBackend
from ... import type_checks

# These classes don't get decorators because these classes are blacklisted by
# the other backend.
CLASS_BLACKLIST = [
    "PntosMediator",
    "PntosKeyValueStore",
    "PntosRegistry",
    "PntosStandardInertialMechanization",
]

class_template = """
py::class_<{class_name}, shared_ptr<{class_name}>>(m, "{class_name}", py::dynamic_attr(){doc_string_param})
{functions};
"""

decorator_template = """
// decorator: A function that accepts a Python object and returns a Python function
auto decorator_{class_name} = [=](py::object plugin_class) {{
    // factory: Constructs a Python object of type `plugin_class` with the
    // Python arguments (`args`) and keyword arguments (`kwargs`)
    // provided.  The Python object is then wrapped into a C struct
    // shared pointer using a Python binding "create_<plugin>"
    // created elsewhere, and that shared pointer is converted into a
    // Python object and returned.
    auto factory = [=](const py::args& args, const py::kwargs& kwargs) {{
        // Call the `plugin_class` variable constructor.
        auto py_plugin_instance = plugin_class(*args, **kwargs);

        // Find the python binding for the C++ function that will create a
        // pntOS C struct wrapped in a shared pointer
        auto py_bind_c_struct_create_function =
            py::module_::import("pntos").attr("create_{class_name}");

        // Call the python binding.  The binding takes the shared pointer
        // return value of the underlying C++ function and converts it into a Python object.
        auto py_c_plugin_instance = py_bind_c_struct_create_function(py_plugin_instance);

        return py_c_plugin_instance;
    }};

    // Return decorator, a function that allows users to pass constructor
    // arguments.
    return py::cpp_function(factory);
}};

// Forward-declare function (will be defined by other backend).
// TODO: use a header?
shared_ptr<{class_name}> create_{class_name} (py::object python_object);
m.def("create_{class_name}", &create_{class_name});

m.def("class_{class_name}", decorator_{class_name});
"""


class PybindPyToC(StructBackend):
    def __init__(self, class_name: str, support_doc_strings=True):
        self.doc_strings = support_doc_strings
        self.class_buf: List[str] = []
        self.class_name = class_name

    def generate(self) -> str:
        functions_str = "\n".join(self.class_buf)

        if self.doc_strings:
            doc_string_param = (
                f", _docstring_class_{self.class_name}_main_class"
            )
        else:
            doc_string_param = ""
        # These classes aren't processed by the C-to-Py backend, so decorators
        # aren't possible.
        if (
            self.class_name in CLASS_BLACKLIST
            or not self.class_name.startswith("Pntos")
        ):
            return class_template.format(
                class_name=self.class_name,
                functions=functions_str,
                doc_string_param=doc_string_param,
            )
        # Everything else gets a decorator.
        else:
            return decorator_template.format(
                class_name=self.class_name
            ) + class_template.format(
                class_name=self.class_name,
                functions=functions_str,
                doc_string_param=doc_string_param,
            )

    def process_func_ptr_field_with_self(
        self,
        field_name: str,
        params: List[Parameter],
        return_t: DecoratedType,
        doc_string: str,
    ):
        # The set of strings that will be joined together in order to form the set
        # of parameters which are the input to the generated Python method
        param_buf = []
        # The set of strings that will be joined together in order to form the
        # parameter conversion portion of the generated Python method body
        body_buf = []
        # The set of strings that will be joined together in order to convert and
        # return the output of the called function pointer back to Python
        return_buf = [""]

        # Add the self param by default, since this func assumes "with_self"
        param_buf += [f"{self.class_name}* self"]

        # The set of strings that are suffixes on the parameter passed into the
        # func pointer call. For example, if the `body_buf` processing a parameter
        # called `foo` needed to make a temporary called `foo_temp` to convert
        # `foo` into something suitable for usage, then this list would contain
        # `_temp` so that the func call will use the `foo_temp` variable instead of
        # the original `foo`.
        param_suffix = []

        # Allow the generators below to signal they are incomplete, such that in
        # the codegen we in form the user we've hit an unsupported use case
        failed_gen = False

        # Extra params
        for param in params[1:]:
            # Special case: char**, interpreted as a NULL terminated array of C
            # strings.

            # While it is not safe in general to assume NULL-termination for
            # this type of parameter, since we are receiving sized arrays from
            # Python and writing out NULL-terminated arrays for C, it is safe
            # for our use case. Worst case scenario, we NULL-terminate when
            # NULL termination wasn't required, and waste a byte.
            if type_name := type_checks.is_string_array(param.type):
                param_buf += [f"std::vector<std::string> {param.name}"]
                param_suffix += ["_prep"]
                body_buf += [
                    "// WIP: TODO: This malloc leaks",
                    f"char** {param.name}_prep ="
                    f" (char**)malloc(({param.name}.size() + 1)*sizeof(char *));",
                    "",
                    f"for (size_t i=0; i < {param.name}.size(); i++) {{",
                    f"    {param.name}_prep[i] ="
                    f" const_cast<char*>({param.name}[i].c_str());",
                    "}",
                    f"{param.name}_prep[{param.name}.size()] = nullptr;",
                    "",
                ]

            # Special case for PntosStringArray
            elif type_name := type_checks.is_pntos_string_array(param.type):
                param_buf += [f"std::vector<std::string> {param.name}"]
                param_suffix += ["_prep"]
                body_buf += [
                    f"char** {param.name}_str_array ="
                    f" (char**)malloc(({param.name}.size() + 1)*sizeof(char *));",
                    "",
                    f"for (size_t i=0; i < {param.name}.size(); i++) {{",
                    f"    {param.name}_str_array[i] ="
                    f" const_cast<char*>({param.name}[i].c_str());",
                    "}",
                    f"{param.name}_str_array[{param.name}.size()] = nullptr;",
                    f"auto {param.name}_prep = pntos_string_array_new({param.name}_str_array);"
                    "",
                ]

            # Special case: char*, interpreted as a C string
            elif type_name := type_checks.is_char_pointer(param.type):
                param_buf += [f"std::string {param.name}"]
                body_buf += [
                    f"char* {param.name}_prep = const_cast<char"
                    f" *>({param.name}.c_str());"
                ]
                param_suffix += ["_prep"]

            # Special case: double*, interpreted as an array of doubles.
            elif type_name := type_checks.is_pointer(param.type, ["double"]):
                param_buf += [f"std::vector<double> {param.name}"]
                body_buf += [
                    f"double* {param.name}_prep = {param.name}.data();"
                ]
                param_suffix += ["_prep"]

            # Special case: PntosMatrix*, because Python has built-in arrays we
            # want to convert to for a better user experience
            elif type_name := type_checks.is_matrix(param.type):
                param_buf += [
                    "py::array_t<double, py::array::c_style |"
                    f" py::array::forcecast> {param.name}"
                ]
                body_buf += [
                    f"PntosMatrix* {param.name}_prep ="
                    f" numpy_to_pntos_matrix({param.name});"
                ]
                param_suffix += ["_prep"]

            # A pointer to a struct starting with Pntos, that is not a special case
            # above
            elif type_name := type_checks.is_pntos_struct_pointer(param.type):
                param_buf += [f"{type_name}* {param.name}"]
                body_buf += [f"pntos_ref({param.name});"]
                param_suffix += [""]

            # A pointer to pointer to Pntos type, interpreted as a NULL terminated
            # array of Pntos struct pointers

            # While it is not safe in general to assume NULL-termination for
            # this type of parameter, since we are receiving sized arrays from
            # Python and writing out NULL-terminated arrays for C, it is safe
            # for our use case. Worst case scenario, we NULL-terminate when
            # NULL termination wasn't required, and waste a byte.
            elif type_name := type_checks.is_pntos_struct_array_pointer(
                param.type
            ):
                param_buf += [
                    f"std::vector<shared_ptr<{type_name}>> {param.name}"
                ]
                param_suffix += ["_prep"]
                body_buf += [
                    "// WIP: TODO: This malloc leaks",
                    f"{type_name}** {param.name}_prep ="
                    f" ({type_name}**)malloc(({param.name}.size() +"
                    f" 1)*sizeof({type_name} *));",
                    "",
                    f"for (size_t i=0; i < {param.name}.size(); i++) {{",
                    f"    pntos_ref({param.name}[i].get());",
                    f"    {param.name}_prep[i] = {param.name}[i].get();",
                    "}",
                    f"{param.name}_prep[{param.name}.size()] = nullptr;",
                    "",
                ]

            # Generic value
            elif type_name := type_checks.is_value(param.type):
                param_buf += [f"{type_name} {param.name}"]
                param_suffix += [""]

            else:
                failed_gen = True
                param_buf += [f"/* WIP: unknown */ void* {param.name}"]
                param_suffix += [""]

        param_names_buf = ["self"]
        for idx, param in enumerate(params[1:]):
            if param.name is not None:
                param_names_buf += [param.name + param_suffix[idx]]
            else:
                failed_gen = True
                param_names_buf = [
                    "/* WIP: Failed generation: Unnamed param */"
                ]
                break

        param_names = ", ".join(param_names_buf)

        if not failed_gen:
            body_buf += ["py::gil_scoped_release gil_release;"]

        if has_return := not type_checks.is_void(return_t):
            body_buf += [f"auto out = self->{field_name}({param_names});"]
        else:
            body_buf += [f"self->{field_name}({param_names});"]

        params_str = ", \n            ".join(param_buf)
        if failed_gen:
            body_buf = [
                "/* WIP: Failed generation: Unknown/unsupported parameters */"
            ]
            for param in params:
                body_buf += [f"(void){param.name};"]
            print(
                f"    Failed parse of {self.class_name}.{field_name}: Unknown/unsupported params."
            )

        body_str = "\n            ".join(body_buf)
        # Process return value
        if has_return and not failed_gen:
            failed_gen = False
            # Special case: char**, interpreted as a NULL terminated array of C
            # strings
            if type_name := type_checks.is_string_array(return_t):
                return_buf += ["/* WIP: todo*/"]
                failed_gen = True

            elif type_name := type_checks.is_pntos_string_array(return_t):
                return_buf += [
                    """
                    std::vector<std::string> out_prep;
                    size_t ii = 0;
                    while (out->data[ii] != nullptr) {{
                        out_prep.push_back(out->data[ii]);
                        ii++;
                    }}
                    return out_prep;
                    """
                ]

            # Special case: char*, interpreted as a C string
            elif type_name := type_checks.is_char_pointer(return_t):
                return_buf += ["return std::string(out);"]

            # Special case: PntosString*
            elif type_name := type_checks.is_string(return_t):
                return_buf += ["return std::string(out->data);"]

            # Special case: PntosMatrix*, because Python has built-in arrays we
            # want to convert to for a better user experience
            elif type_name := type_checks.is_matrix(return_t):
                return_buf += ["return pntos_matrix_to_numpy(out);"]

            # Special case: PntosDoubleArray*
            elif type_name := type_checks.is_pntos_double_array(return_t):
                return_buf += ["return pntos_double_array_to_numpy(out);"]

            # A pointer to a struct starting with Pntos, that is not a special case
            # above
            elif type_name := type_checks.is_pntos_struct_pointer(return_t):
                return_buf += [
                    f"return shared_ptr<{type_name}>(out, pntos_deleter);"
                ]

            # Generic value
            elif type_name := type_checks.is_value(return_t):
                return_buf += ["return out;"]

            else:
                failed_gen = True

            if failed_gen:
                return_buf += ["/* WIP: Failed return value generation */\n"]
                return_buf += ["(void)out;"]
                print(
                    f"    Failed return value generation for {self.class_name}.{field_name}:"
                )

            return_str = "\n            ".join(return_buf)
        else:
            return_str = ""

        # Generate param name annotations for pybind11 to know the names of params
        # (e.g. `py::arg("foo")`)
        py_arg_buf = [f"""py::arg("{param.name}")""" for param in params[1:]]

        if py_arg_buf == []:
            py_args = ""
        else:
            py_args = ",\n    " + ", ".join(py_arg_buf)

        if self.doc_strings:
            doc_string_param = f""",
    _docstring_class_{self.class_name}_{field_name}"""
        else:
            doc_string_param = ""

        out = f"""
         .def("{field_name}",
    []({ params_str }) {{

            { body_str }{ return_str }

    }}{doc_string_param}{py_args})
        """.strip()
        self.class_buf += [out]

    def process_data_pointer_field(
        self,
        field_name: str,
        type_name: str,
        data_len: Union[str, int],
        doc_string: str,
        deref="",
    ):
        class_name = self.class_name
        # When we have a field like `double (*foo)[4];`, we need to make sure
        # we deref in the correct order. In particular, if we did:
        #     out[i] = *self->foo[i];
        # then the `[i]` operator binds tighter than `*`, meaning we will be
        # iterating over length 4 arrays instead of elements, which is off the
        # end of the allocated memory. Thus, we need to put parens around our
        # deref to force the deref to occur first, _then_ we iterate over
        # elements in the size-4 array. While it is safe to always have these
        # parents present, it adds noise to other data pointer fields, so we'll
        # make our generator code a little bit messy in order to make our
        # output clean.
        if deref != "":
            open_paren = "("
            close_paren = ")"
        else:
            open_paren = ""
            close_paren = ""
        self.class_buf += [
            f""".def_property_readonly("{field_name}", []({class_name}* self){{
            std::vector<{type_name}> out({data_len});
            for (decltype({data_len}) i=0; i < {data_len}; i++) {{
                out[i] = {open_paren}{deref}self->{field_name}{close_paren}[i];
            }}
            return out;
        }})"""
        ]

    def process_outer_managed_pointer_field(
        self, field_name: str, type_name: str, doc_string: str
    ):
        class_name = self.class_name

        if type_name == 'PntosMatrix':
            self.class_buf += [
                f""".def_property_readonly("{field_name}", []({class_name}* self) {{
                return pntos_matrix_to_numpy(self->{field_name});
            }})"""
            ]
        else:
            self.class_buf += [
                f""".def_property_readonly("{field_name}", []({class_name}* self) {{
                return shared_ptr<{type_name}>(self->{field_name},
                    captured_pntos_deleter_generator(self));
            }})"""
            ]

    def process_outer_managed_pointer_array_field(
        self,
        field_name: str,
        field_type_name: str,
        data_len: Union[str, int],
        doc_string: str,
        deref="",
    ):
        class_name = self.class_name
        self.class_buf += [
            f""".def_property_readonly("{field_name}", []({class_name}* self) {{
            std::vector<shared_ptr<{field_type_name}>> out({data_len});
            for (decltype({data_len}) i=0; i < {data_len}; i++) {{
                out[i] = shared_ptr<{field_type_name}>(self->{field_name}[i],
                    pntos_deleter);
            }}
            return out;
        }})"""
        ]

    def process_string_field(self, field_name: str, doc_string: str):
        class_name = self.class_name
        self.class_buf += [
            f""".def_property_readonly("{field_name}", []({class_name}* self){{
            return std::string(self->{field_name});
        }}, _docstring_class_{class_name}_{field_name})"""
        ]

    def process_string_array_field(self, field_name: str, doc_string: str):
        class_name = self.class_name
        self.class_buf += [
            f""".def_property_readonly("{field_name}", []({class_name}* self){{
            std::vector<std::string> out;
            for (int i=0; self->{field_name}[i] != nullptr; i++) {{
                out.emplace_back(std::string(self->{field_name}[i]));
            }}
            return out;
        }})"""
        ]

    def process_simple_field(
        self, field_name: str, field_type_name: str, doc_string: str
    ):
        class_name = self.class_name
        self.class_buf += [
            f""".def_readonly("{field_name}", &{class_name}::{field_name})"""
        ]

    def process_pntos_inheritance_field(
        self, field_name: str, field_type_name: str, doc_string: str
    ):
        class_name = self.class_name
        self.class_buf += [
            f""".def_property_readonly("{field_name}", []({class_name}* self) {{
            {field_type_name}* out = &self->{field_name};
            pntos_ref(self);
            return shared_ptr<{field_type_name}>(out, pntos_deleter);
        }})"""
        ]

    def process_class_docstring(self, doc_string: str):
        pass
