from abc import ABCMeta
import os
import re
from subprocess import check_call

from typing import List, Dict
import cxxheaderparser.simple as p
from cxxheaderparser.simple import ClassScope
from cxxheaderparser.simple import NamespaceScope
from cxxheaderparser.types import (
    Field,
    AnonymousName,
    NameSpecifier,
    DecoratedType,
    Pointer,
    FunctionType,
    Typedef,
)

from firehose import type_checks
from firehose.backends import (
    StructBackend,
    PybindCToPy,
    PybindPyToC,
    DocstringExtractor,
)

import argparse

parser = argparse.ArgumentParser()
parser.add_argument(
    "project_root", help="Path to project root for input files"
)
parser.add_argument("output_folder", help="Path to folder for output files")
parser.add_argument(
    "--project",
    help="Which project to operate on, defaults to pntos",
    choices=["pntos", "navtk"],
    default="pntos",
)
args = parser.parse_args()


#  TODO: CLI params?
CLASS_BLACKLIST = [
    "PntosPluginsDescription",
    "PntosAspnStreamConfiguration",
    "PntosInertialState",
    "PntosMatrix",
]


pybind_template = """
# include <pntos.h>
# include <pntos/utils.h>
# include <memory>
# include <pybind11/stl.h>
# include <pybind11/numpy.h>
# include "deleters.h"
# include "utility_functions.h"
# include "custom_bindings.h"
# include "generated_docs.h"
# include "utility_functions.h"
namespace py = pybind11;
using namespace std;

{utilities}

PYBIND11_MODULE(pntos, m) {{
    m.doc() = "pntOS python bindings module";

    // Insert hand-rolled bindings into the module first.
    add_custom_definitions(m);
    add_utility_definitions(m);

    {definitions}
}}
"""


pybind_trampolines_template = """
#include <pntos.h>
#include <pntos/utils.h>
#include <memory>
#include <Python.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>
#include <pybind11/numpy.h>
#include "deleters.h"
#include "utility_functions.h"

namespace py=pybind11;
using namespace std;

{definitions}
"""

function_typedefs: Dict[str, Typedef] = {}


def should_process(name):
    return name not in CLASS_BLACKLIST


def gen_namespace_classes(
    namespace: NamespaceScope, generator_factory: ABCMeta
) -> str:
    out_buf = []

    # Record all function typedefs in this namespace
    for typedef in namespace.typedefs:
        if isinstance(typedef.type, Pointer):
            if isinstance(typedef.type.ptr_to, FunctionType):
                function_typedefs[typedef.name] = typedef

    for class_ in namespace.classes:
        first_segment = class_.class_decl.typename.segments[0]
        if isinstance(first_segment, AnonymousName):
            raise RuntimeError("No support for anonymous structs")

        elif isinstance(first_segment, NameSpecifier):
            name = first_segment.name
            print(f"Processing {name} with {generator_factory.__name__}")
            if should_process(name):
                # Clean up ASPN's mess
                if name.startswith("Spicy"):
                    class_name = "Aspn3" + name[5:]
                else:
                    class_name = name

                class_out = gen_class(
                    class_, name, generator_factory(class_name)
                )
                out_buf += [class_out]
        else:
            raise RuntimeError("Encountered class with unknown structure")
    out = "\n".join(out_buf)
    return out


def acceptable_input_file(filename: str):
    return (
        not filename.endswith(".c")
        and not filename.endswith(".cpp")
        and not filename.endswith(".build")
        and os.path.isfile(filename)
    )


def parse_file(in_file: str):
    print(f"Reading {in_file}")
    with open(in_file, "r", encoding="utf-8") as f_handle:
        f_string = f_handle.read()
        f_string = re.sub(
            "PNTOS_NULLABLE|PNTOS_ASSUME_NONNULL_END|PNTOS_ASSUME_NONNULL_BEGIN|SPICY_CXX_BOILERPLATE_TOP|SPICY_CXX_BOILERPLATE_END",
            "",
            f_string,
        )
        f_string = re.sub(
            r"#ifdef PNTOS_STATIC_BUILD.*#endif", "", f_string, flags=re.DOTALL
        )
        out = p.parse_string(f_string)
        return out.namespace


# Array fields have a length field in the same structure. Search for a length
# for target_field among all_fields and return the length field name.  Not
# finding the field length name terminates the script, requiring additional
# special cases to be implemented.
def find_arraylength(target_field: Field, all_fields: List[Field]) -> str:
    data_len = None

    # Most lengths are defined by a single other field
    for field_candidate in all_fields:

        # Some ASPN array fields have a length that is the same name with `num_` prepended
        # (mostly `error_model_params` and `num_error_model_params`)
        if field_candidate.name == f"num_{target_field.name}":
            data_len = "self->" + field_candidate.name

        # Many ASPN covariance arrays have a companion length field `num_meas`
        elif (
            field_candidate.name == "num_meas"
            and target_field.name == "covariance"
        ):
            data_len = "self->num_meas*self->num_meas"

        # Special case for PntosDoubleArray
        elif (
            field_candidate.name == "num_doubles"
            and target_field.name == "data"
        ):
            data_len = "self->num_doubles"

        # Special case for PntosIntArray
        elif (
            field_candidate.name == "num_ints" and target_field.name == "data"
        ):
            data_len = "self->num_ints"

        # Special case for PntosMessageArray
        elif (
            field_candidate.name == "num_messages"
            and target_field.name == "data"
        ):
            data_len = "self->num_messages"

        # Special case for Aspn3MeasurementGNSSSubframe
        elif (
            field_candidate.name == "num_bytes"
            and target_field.name == "data_vector"
        ):
            data_len = "self->num_bytes"

        # Special case for Aspn3MeasurementTimeAbsolute
        elif field_candidate.name == "num_obs" and target_field.name in [
            "clock_id",
            "elapsed_sec",
            "elapsed_nsec",
            "elapsed_attosec",
            "covariance",
        ]:
            data_len = "self->num_obs"

        # Special case for Aspn3MeasurementGNSSWithSvData
        elif (
            field_candidate.name == "num_signals_tracked"
            and target_field.name == "obs"
        ):
            data_len = "self->num_signals_tracked"

        # Special case for Aspn3MeasurementGNSS* classes
        elif (
            field_candidate.name == "num_signals_tracked"
            and target_field.name == "sv_data"
        ):
            data_len = "self->num_signals_tracked"

        # Special case for Aspn3MeasurementDirection2dToPointKnown
        elif (
            field_candidate.name == "num_obs"
            and target_field.name == "point_position_and_covariance"
        ):
            data_len = "self->num_obs"

        # Special case for Aspn3Image
        elif (
            field_candidate.name == "image_data_length"
            and target_field.name == "image_data"
        ):
            data_len = "self->image_data_length"

        # Special case for PntosCrossCovariances
        elif (
            field_candidate.name == "num_covariances"
            and target_field.name == "cross_covariances"
        ):
            data_len = "self->num_covariances"

    # Some lengths are defined by a combination of other fields

    # Special case for PntosMatrix
    if not data_len:
        field_names = [target_field.name for target_field in all_fields]
        if (
            target_field.name == "data"
            and "rows" in field_names
            and "cols" in field_names
        ):
            data_len = "self->rows * self->cols"

    if not data_len:
        raise RuntimeError(
            f"Special case required to find length for {target_field.name}"
        )

    return data_len


def gen_class(
    class_: ClassScope, class_name: str, class_generator: StructBackend
) -> str:

    if not (class_doc_string := class_.class_decl.doxygen):
        class_doc_string = "<Missing Docstring>"

    class_generator.process_class_docstring(class_doc_string)

    for field_num, field in enumerate(class_.fields):
        field_name: str = str(field.name)
        if not (doc_string := field.doxygen):
            doc_string = "<Missing Docstring>"
        field_type: DecoratedType = field.type

        # Special case: typedef of a function with self as first parameter
        if func_info := type_checks.is_function_typedef_with_self(
            field_type, class_name, function_typedefs
        ):
            (params, return_type) = func_info
            class_generator.process_func_ptr_field_with_self(
                field_name, params, return_type, doc_string
            )

        # `char**`, interpreted as a NULL terminated array of strings
        elif type_checks.is_string_array(field_type):
            class_generator.process_string_array_field(field_name, doc_string)

        # Special case: do not copy pntOS inheritance structs, but grab a
        # reinterpreted ref instead. Requires: first field inheritance, outer
        # struct can pntos_ref, inner struct can pntos_unref.
        elif (
            field_name == "common"
            and type_checks.is_value(field_type)
            and field_num == 0
            and type_checks.is_value(field_type).startswith("Pntos")
            and class_name.startswith("Pntos")
        ):
            type_name = type_checks.is_value(field_type)
            class_generator.process_pntos_inheritance_field(
                field_name, type_name, doc_string
            )

        # Other fields which are simple values (non-pointer)
        elif type_name := type_checks.is_value(field_type):
            class_generator.process_simple_field(
                field_name, type_name, doc_string
            )

        # Field is function pointer with self
        elif func_info := type_checks.is_function_pointer_with_self(
            field_type, class_name
        ):
            (params, return_type) = func_info
            class_generator.process_func_ptr_field_with_self(
                field_name, params, return_type, doc_string
            )

        # Field is function pointer without a self reference
        elif func_info := type_checks.is_function_pointer(field_type):
            (params, return_type) = func_info
            raise RuntimeError(
                f"Function pointer without self reference unsupported ({field.name})"
            )

        # Special case: `char *`, interpreted as a C string
        elif type_checks.is_char_pointer(field_type):
            class_generator.process_string_field(field_name, doc_string)

        # Special case: `double[N]`
        elif data_len := type_checks.is_double_array(field_type):
            class_generator.process_data_pointer_field(
                field_name, "double", int(data_len), doc_string
            )

        # Special case: `float[N]`
        elif data_len := type_checks.is_float_array(field_type):
            class_generator.process_data_pointer_field(
                field_name, "float", int(data_len), doc_string
            )

        # Special case: `<int>[N]` for various kinds of <int>
        elif array_info := type_checks.is_int_array(field_type):
            (element_type, data_len) = array_info
            class_generator.process_data_pointer_field(
                field_name, element_type, int(data_len), doc_string
            )

        # Special case: `double (*)[N]` (That is: pointer to array of size
        # N doubles)
        elif data_len := type_checks.is_pointer_to_double_array(field_type):
            class_generator.process_data_pointer_field(
                field_name, "double", int(data_len), doc_string, deref="*"
            )

        # Special case: `<int> *` or `double *`, interpreted as an array of
        # types of integers or doubles
        elif type_name := type_checks.is_pointer(
            field_type,
            ["double", "int8_t", "int64_t", "uint8_t", "uint32_t", "uint64_t"],
        ):
            data_len = find_arraylength(field, class_.fields)
            class_generator.process_data_pointer_field(
                field_name, type_name, data_len, doc_string
            )

        # Special case: Pointer to field type that starts with `Aspn3Type`,
        # interpreted as an array of that type
        elif (
            type_name := type_checks.is_pointer(field_type)
        ) and field_type.ptr_to.typename.segments[0].name.startswith(
            "Aspn3Type"
        ):
            data_len = find_arraylength(field, class_.fields)
            class_generator.process_data_pointer_field(
                field_name, type_name, data_len, doc_string
            )

        # Void pointers unsupported
        elif type_name := type_checks.is_pointer(field_type, ["void"]):
            raise RuntimeError(
                f"void pointer fields like {field.name} are not supported"
            )

        # Intentionally do not wrap memory managers (Python refcounting is
        # implicitly handled)
        elif type_name := type_checks.is_pointer(
            field_type, ["PntosManagedMemory"]
        ):
            pass

        # Pointer to a type other than the special cases above, managed by
        # a Pntos class
        elif (
            type_name := type_checks.is_pointer(field_type)
        ) and class_name.startswith("Pntos"):
            class_generator.process_outer_managed_pointer_field(
                field_name, type_name, doc_string
            )

        # Pointer to a pointer to a type other than the special cases above,
        # managed by a Pntos class
        elif (
            type_name := type_checks.is_pointer_to_pointer(field_type)
        ) and class_name.startswith("Pntos"):

            data_len = find_arraylength(field, class_.fields)
            class_generator.process_outer_managed_pointer_array_field(
                field_name, type_name, data_len, doc_string
            )
        else:
            raise RuntimeError(f"The field {field.name} was not handled")

    return class_generator.generate()


def gen_enum(enum_name: str, enum_values: List[str]) -> str:
    out_buf = []
    out_buf += [f"""    py::enum_<{enum_name}>(m, "{enum_name}")"""]
    for enum_value in enum_values:
        out_buf += [f""".value("{enum_value}", {enum_name}::{enum_value})"""]
    out_buf += [";"]
    out = "\n    ".join(out_buf)
    return out


if args.project == "pntos":
    search_roots = [
        f"{args.project_root}/api/include/pntos/plugins",
        f"{args.project_root}/subprojects/aspn-c/src/aspn3",
        f"{args.project_root}/subprojects/libspicyc/src/libspicyc",
    ]
    individual_files = [f"{args.project_root}/api/include/pntos/loader.h"]
    exclude_files = ["helper_macros.h"]
elif args.project == "navtk":
    search_roots = [f"{args.project_root}/subprojects/navtoolkit/src/navtk"]
    individual_files = []
    exclude_files = [
        "not_null.hpp",
        "factory.hpp",
        "filesystem.hpp",
        "GpsEphemerisHandler.hpp",
    ]
else:
    raise RuntimeError(f"Project type {args.project} not implemented")

input_files = individual_files
for folder in search_roots:
    print(f"Looking in {folder}")
    for root, dirs, files in os.walk(folder):
        candidates = [
            f"{root}/{file}" for file in files if file not in exclude_files
        ]
        input_files += filter(acceptable_input_file, candidates)


def enum_to_class(enum_name: str) -> str:
    """
    Accepts a string to a PntosAspnCombinedType enum value and returns the
    equivalent class name. For example, an input of
    'ASPN3_MEASUREMENT_POSITION' will return 'Aspn3MeasurementPosition'.
    """
    out = ''
    pieces = enum_name.split('_')
    for piece in pieces:
        # Special case: these defy convention and should remain upper case.
        if piece in ['GNSS', 'GPS', 'IMU', 'UTC', 'INS', 'TDOA']:
            out += piece
        # Special case: 'GNSS_M' defies convention; should remain upper case
        # and be concatenated with previous chunk.
        elif piece == 'M':
            out += piece
        # Special case: if it starts with a number then the next character is a
        # capital letter, *unless* it's a D, like '3D'.
        elif piece[0].isdigit() and piece[1] != 'D':
            out += piece[0:2] + piece[2:].lower()
        # Default case: PascalCase.
        else:
            out += piece[0] + piece[1:].lower()
    return out


def gen_bindings():
    # This retains all files, so not suitable for very large inputs
    asts = {}
    for input_file in input_files:
        ast = parse_file(input_file)
        asts[input_file] = ast

    pybind_list = []
    pybind_utilities_list = []
    pybind_trampolines_list = []
    doc_list = []

    for input_file, ast in asts.items():
        namespace = ast
        doc = gen_namespace_classes(namespace, DocstringExtractor)
        pybind = gen_namespace_classes(namespace, PybindPyToC)
        pybind_trampolines = gen_namespace_classes(namespace, PybindCToPy)
        # Hard-coded pybind11 enum handling for now
        enum_out_buf = []
        for enum in namespace.enums:
            first_segment = enum.typename.segments[0]
            if isinstance(first_segment, NameSpecifier):
                enum_name = first_segment.name
                enum_values = enum.values
                enum_out_buf += [
                    gen_enum(enum_name, [val.name for val in enum_values])
                ]
            else:
                raise RuntimeError("Encountered enum with unknown structure")

            enum_name = first_segment.name

            # Create downcast_message function from this enum.
            if enum_name == 'PntosAspnCombinedType':
                downcast_message_func = [
                    """\
                    /**
                    * Downcast an ASPN message to the type stored in the message, and then
                    * cast it to a Python object and return it.
                    *
                    * If pntos_message is not an ASPN message, the return will be py::none().
                    *
                    * If pntos_message is a nullptr or its wrapped_message member is a nullptr,
                    * the return will be py::none().
                    *
                    * The returned object holds a reference to the pntos_message argument.
                    * This function will increment the pntos_message argument reference count.
                    * Once the return object is destroyed, the reference count on pntos_message
                    * will be decremented.
                    */
                    py::object downcast_aspn_message(std::shared_ptr<PntosMessage> pntos_message) {

                    if (pntos_message == nullptr) {
                        return py::none();
                    }

                    if (pntos_message->wrapped_message == nullptr) {
                        return py::none();
                    }

                    if (pntos_message->is_aspn_message == false) {
                        return py::none();
                    }

                    auto aspn_object = reinterpret_cast<Aspn3Object*>(pntos_message->wrapped_message);

                    switch (aspn_object->runtime_type) {"""
                ]
                for enum_value in enum_values:
                    # Skip enum values that don't correspond to an ASPN type.
                    if enum_value.value is None:
                        continue
                    token = enum_value.value.tokens[0]

                    # Skip enum values that don't correspond to an ASPN type.
                    if token.type != 'NAME':
                        continue

                    name = token.value

                    # Convert 'ASPN3_MEASUREMENT_FOO' to
                    # 'Aspn3MeasurementFoo'.
                    aspn_class = enum_to_class(name)
                    downcast_message_func += [
                        f"""\
                    	case {name}: {{
                            auto aspn_measurement = reinterpret_cast<{aspn_class}*>(aspn_object);
                            auto aspn_shared_ptr = shared_ptr<{aspn_class}>(aspn_measurement,
                                captured_pntos_deleter_generator(pntos_message.get()));
                            return py::cast(aspn_shared_ptr);
                        }} break;
                        """
                    ]
                downcast_message_func += [
                    """default: {
                        printf("Could not downcast ASPN message: No match found for %d\\n", aspn_object->runtime_type);
                        // Couldn't find a match, abort downcast.
                        return py::cast(aspn_object);
                    }
                    }
                    }"""
                ]
                pybind_utilities_list += ['\n'.join(downcast_message_func)]
                pybind_list += [
                    'm.def("downcast", &downcast_aspn_message, py::arg("aspn_message"));'
                ]

            # Generate and bind downcast functions for pntOS enums.

            # Blacklist enums that don't serve as downcast guiders.
            elif 'Pntos' in enum_name and enum_name not in [
                'PntosSolutionType',
                'PntosLoggingLevel',
                'PntosKeyValueStoreDataFormat',
                'PntosAspnExtensionType',
                'PntosInertialSolutionRangeType',
                'PntosFusionType',
            ]:
                # This dictionary, given the name of an enum, returns an array
                # corresponding to the relevant class. For example,
                #  PntosPluginTypes is an emum that is used to downcast
                # PntosCommonPlugin and is stored on that class' plugin_type
                # field.
                class_enum_map = {
                    'PntosPluginTypes': ['PntosCommonPlugin', 'plugin_type'],
                    'PntosFusionType': [
                        'PntosCommonFusionEngine',
                        'engine_type',
                    ],
                    'PntosMechanizationType': [
                        'PntosCommonInertialMechanization',
                        'mechanization_type',
                    ],
                }
                class_name = class_enum_map[enum_name][0]

                field_name = class_enum_map[enum_name][1]
                downcast_message_func = [
                    f"""
                    py::object downcast_{class_name}(std::shared_ptr<{class_name}> object) {{
                        switch (object->{field_name}) {{
                    """
                ]

                for value in enum.values:
                    name = value.name
                    corresponding_class = enum_to_class(name)

                    # Special case: nonstandard mapping (#591)
                    if corresponding_class in [
                        'PntosStandardMechanizationModel',
                        'PntosExternalMechanizationModel',
                    ]:
                        corresponding_class = (
                            'PntosStandardInertialMechanization'
                        )

                    # Blacklisted enum values; these classes aren't defined.
                    if name in [
                        'PNTOS_UNDEFINED_PLUGIN',
                        'PNTOS_ALIGNMENT_PLUGIN',
                        'PNTOS_DATABASE_PLUGIN',
                        'PNTOS_ORCHESTRATION_STRATEGY_PLUGIN',
                        'PNTOS_UTILITY_PLUGIN',
                        'PNTOS_DATA_PREPROCESSOR_PLUGIN',
                        'PNTOS_NUM_PLUGIN_TYPES',
                        'PNTOS_NON_ORTHOGONAL_MECHANIZATION_MODEL',
                    ]:
                        downcast_message_func += [
                            f"""\
                        case {name}: {{
                            printf("No such class, {corresponding_class}\\n");
                            py::object out = py::cast(object);
                            return out;
                        }}
                        """
                        ]
                        continue
                    downcast_message_func += [
                        f"""\
                    case {name}: {{
                        pntos_ref(object.get());
                        shared_ptr<{corresponding_class}> child = shared_ptr<{corresponding_class}>(
                            ({corresponding_class}*)(object.get()), pntos_deleter);
                        printf("Found {corresponding_class}\\n");
                        py::object out = py::cast(child);
                        return out;
                    }}
                    """
                    ]
                downcast_message_func += [
                    """\
                default: {{
                        printf("No match found\\n");
                        // Couldn't find a match, abort downcast and return parent class.
                        return py::cast(object);
                    }}
                }}
                """
                ]
                pybind_utilities_list += ['\n'.join(downcast_message_func)]
                pybind_list += [
                    f'm.def("downcast", &downcast_{class_name}, py::arg("aspn_message"));'
                ]

        pybind += "\n" + "\n".join(enum_out_buf)

        doc_list += [doc]
        pybind_list += [pybind]
        pybind_trampolines_list += [pybind_trampolines]

    pybind_string = pybind_template.format(
        utilities="\n".join(pybind_utilities_list),
        definitions="\n".join(pybind_list),
    )
    pybind_trampolines_string = pybind_trampolines_template.format(
        definitions="\n".join(pybind_trampolines_list)
    )
    doc_string = "\n".join(doc_list)

    with open(
        f"{args.output_folder}/generated_pybind.cpp", "w", encoding="utf-8"
    ) as file:
        file.write(pybind_string)

    with open(
        f"{args.output_folder}/generated_docs.h", "w", encoding="utf-8"
    ) as file:
        file.write(doc_string)

    with open(
        f"{args.output_folder}/generated_pybind_adapters.cpp",
        "w",
        encoding="utf-8",
    ) as file:
        file.write(pybind_trampolines_string)

    try:
        check_call(
            [
                'clang-format',
                '-style=file',
                '--verbose',
                '-i',
                f"{args.output_folder}/generated_pybind.cpp",
                f"{args.output_folder}/generated_docs.h",
                f"{args.output_folder}/generated_pybind_adapters.cpp",
            ]
        )
    except FileNotFoundError:
        print('Could not find clang-format; unable to format output files.')


if __name__ == "__main__":
    gen_bindings()
